#ifndef MANAGESIEVE_LOGIN_SETTINGS_PLUGIN_H
#define MANAGESIEVE_LOGIN_SETTINGS_PLUGIN_H

#include "lib.h"

void managesieve_login_settings_init(struct module *module);
void managesieve_login_settings_deinit(void);

#endif
